import asyncio
from unittest import mock

import pytest
import pytest_asyncio
from hllrcon.connection import RconConnection
from hllrcon.exceptions import HLLConnectionLostError
from hllrcon.protocol.protocol import RconProtocol
from hllrcon.protocol.response import RconResponse, RconResponseStatus


@pytest.fixture
def protocol() -> RconProtocol:
    """Fixture to create a mock RconProtocol."""
    mock_protocol = mock.Mock(spec=RconProtocol)

    def connection_lost(exc: Exception | None) -> None:
        """Mock connection lost method."""
        mock_protocol.is_connected.return_value = False
        mock_protocol.on_connection_lost(exc)

    mock_protocol.is_connected.return_value = True
    mock_protocol.connection_lost = connection_lost
    return mock_protocol


@pytest_asyncio.fixture
async def connection(
    monkeypatch: pytest.MonkeyPatch,
    protocol: RconProtocol,
) -> RconConnection:
    """Fixture to create a mock RconProtocol."""
    monkeypatch.setattr(RconProtocol, "connect", mock.AsyncMock(return_value=protocol))
    return await RconConnection.connect("localhost", 1234, "password")


@pytest.mark.asyncio
async def test_is_connected(connection: RconConnection, protocol: RconProtocol) -> None:
    assert connection.is_connected() is True

    protocol.connection_lost(None)
    assert connection.is_connected() is False


@pytest.mark.asyncio
async def test_disconnect(connection: RconConnection, protocol: mock.Mock) -> None:
    with pytest.raises(asyncio.TimeoutError):
        async with asyncio.timeout(0.1):
            await connection.wait_until_disconnected()

    connection.disconnect()

    protocol.disconnect.assert_called_once()
    protocol.connection_lost(None)

    async with asyncio.timeout(0.1):
        await connection.wait_until_disconnected()


@pytest.mark.asyncio
async def test_execute(
    connection: RconConnection,
    protocol: mock.Mock,
) -> None:
    command = "test_command"
    version = 1
    body = "test_body"
    response = "response"

    protocol.execute.return_value = RconResponse(
        request_id=1,
        command=command,
        version=version,
        status_code=RconResponseStatus.OK,
        status_message="OK",
        content_body="response",
    )

    result = await connection.execute(command, version, body)
    assert result == response

    protocol.connection_lost(None)
    with pytest.raises(HLLConnectionLostError):
        await connection.execute(command, version, body)



# mypy: disable-error-code="prop-decorator"
# ruff: noqa: N802

import re
from enum import StrEnum
from functools import cached_property
from typing import Self

from pydantic import computed_field, model_validator

from hllrcon.data.sectors import (
    SECTORS_CARENTAN_LARGE,
    SECTORS_CARENTAN_SMALL,
    SECTORS_DRIEL_LARGE,
    SECTORS_DRIEL_SMALL,
    SECTORS_ELALAMEIN_LARGE,
    SECTORS_ELALAMEIN_SMALL,
    SECTORS_ELSENBORNRIDGE_LARGE,
    SECTORS_ELSENBORNRIDGE_SMALL,
    SECTORS_FOY_LARGE,
    SECTORS_HILL400_LARGE,
    SECTORS_HILL400_SMALL,
    SECTORS_HURTGENFOREST_LARGE,
    SECTORS_KHARKOV_LARGE,
    SECTORS_KURSK_LARGE,
    SECTORS_MORTAIN_LARGE,
    SECTORS_MORTAIN_SMALL,
    SECTORS_OMAHABEACH_LARGE,
    SECTORS_PURPLEHEARTLANE_LARGE,
    SECTORS_PURPLEHEARTLANE_SMALL,
    SECTORS_REMAGEN_LARGE,
    SECTORS_STALINGRAD_LARGE,
    SECTORS_STALINGRAD_SMALL,
    SECTORS_STMARIEDUMONT_LARGE,
    SECTORS_STMARIEDUMONT_SMALL,
    SECTORS_STMEREEGLISE_LARGE,
    SECTORS_STMEREEGLISE_SMALL,
    SECTORS_TOBRUK_LARGE,
    SECTORS_TOBRUK_SMALL,
    SECTORS_UTAHBEACH_LARGE,
    Grid,
    Sector,
    Strongpoint,
)

from ._utils import (
    CaseInsensitiveIndexedBaseModel,
    class_cached_property,
)
from .factions import Faction
from .game_modes import GameMode
from .maps import Map, Orientation
from .teams import Team

__all__ = (
    "Layer",
    "TimeOfDay",
    "Weather",
)


RE_LAYER_NAME_SMALL1 = re.compile(
    r"^(?P<tag>[A-Z]{3,5})_S_(?P<year>\d{4})_P_(?P<game_mode>\w+?)(?:_(?P<environment>\w+))?$",
)
RE_LAYER_NAME_SMALL2 = re.compile(
    r"^(?P<tag>[A-Z]{3,5})_S_(?P<year>\d{4})_(?:(?P<environment>\w+)_)?P_(?P<game_mode>\w+)$",
)
RE_LAYER_NAME_LARGE = re.compile(
    r"^(?P<tag>[A-Z]{3,5})_L_(?P<year>\d{4})_(?P<game_mode>\w+?)(?P<attackers>[A-Z]\w*?)?(?:_(?P<environment>\w+))?$",
)
RE_LEGACY_LAYER_NAME = re.compile(
    r"^(?P<name>[a-z0-9]+)_(?P<game_mode>(?!off)[a-z]+)(?:_V2)?(?:_(?P<environment>[a-z]+))?$",
)
RE_LEGACY_LAYER_NAME_OFFENSIVE = re.compile(
    r"^(?P<name>[a-z0-9]+)_(?P<game_mode>off(?:ensive)?)_?(?P<attackers>[a-zA-Z]+)(?:_(?P<environment>\w+))?$",
)


class TimeOfDay(StrEnum):
    DAWN = "Dawn"
    DAY = "Day"
    DUSK = "Dusk"
    NIGHT = "Night"


class Weather(StrEnum):
    CLEAR = "Clear"
    OVERCAST = "Overcast"
    RAIN = "Rain"
    SNOW = "Snow"


LAYER_GAME_MODE_MAP: dict[str, GameMode] = {
    "warfare": GameMode.WARFARE,
    "offensive": GameMode.OFFENSIVE,
    "off": GameMode.OFFENSIVE,
    "skirmish": GameMode.SKIRMISH,
}

LAYER_ATTACKERS_MAP: dict[str, Faction] = {
    "us": Faction.US,
    "ger": Faction.GER,
    "dak": Faction.DAK,
    "sov": Faction.SOV,
    "soviet": Faction.SOV,
    "rus": Faction.SOV,
    "russian": Faction.SOV,
    "ussr": Faction.SOV,
    "cw": Faction.CW,
    "gb": Faction.CW,
    "com": Faction.CW,
    "brit": Faction.CW,
    "british": Faction.CW,
    "b8a": Faction.CW,
}

LAYER_ENVIRONMENT_MAP: dict[str, tuple[TimeOfDay, Weather]] = {
    "day": (TimeOfDay.DAY, Weather.CLEAR),
    "dawn": (TimeOfDay.DAWN, Weather.CLEAR),
    "morning": (TimeOfDay.DAWN, Weather.CLEAR),
    "dusk": (TimeOfDay.DUSK, Weather.CLEAR),
    "evening": (TimeOfDay.DUSK, Weather.CLEAR),
    "night": (TimeOfDay.NIGHT, Weather.CLEAR),
    "clear": (TimeOfDay.DAY, Weather.CLEAR),
    "overcast": (TimeOfDay.DAY, Weather.OVERCAST),
    "rain": (TimeOfDay.DAY, Weather.RAIN),
    "snow": (TimeOfDay.DAY, Weather.SNOW),
}


class Layer(CaseInsensitiveIndexedBaseModel):
    map: Map
    game_mode: GameMode
    time_of_day: TimeOfDay
    weather: Weather
    grid: Grid
    sectors: list[Sector]
    attacking_team: Team | None = None

    @model_validator(mode="after")
    def _set_sectors_layer_backref(self) -> Self:
        for sector in self.sectors:
            sector._layer = self  # type: ignore[misc] # noqa: SLF001

            for capture_zone in sector.capture_zones:
                capture_zone._layer = self  # type: ignore[misc] # noqa: SLF001

        return self

    @model_validator(mode="after")
    def _apply_offset_to_strongpoints(self) -> Self:
        for sector in self.sectors:
            for capture_zone in sector.capture_zones:
                # Bit of a hack to get around the model being frozen
                object.__setattr__(
                    capture_zone.strongpoint,
                    "center",
                    (
                        capture_zone.strongpoint.center[0] + self.grid.offset[0],
                        capture_zone.strongpoint.center[1] + self.grid.offset[1],
                        capture_zone.strongpoint.center[2],
                    ),
                )
        return self

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id!r}, map={self.map!r},"
            f" attackers={self.attacking_team!r}, time_of_day={self.time_of_day!r},"
            f" weather={self.weather!r})"
        )

    def __hash__(self) -> int:
        return hash(self.id.lower())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (Layer, str)):
            return str(self).lower() == str(other).lower()
        return NotImplemented

    @classmethod
    def _parse_id(cls, id_: str) -> "Layer":
        exc = ValueError(f"Could not parse layer ID: {id_}")

        for pattern in (
            RE_LAYER_NAME_SMALL1,
            RE_LAYER_NAME_SMALL2,
            RE_LAYER_NAME_LARGE,
            RE_LEGACY_LAYER_NAME,
            RE_LEGACY_LAYER_NAME_OFFENSIVE,
        ):
            if match := pattern.match(id_):
                break
        else:
            raise exc

        groups = match.groupdict("")

        map_id: str
        map_tag: str
        if groups.get("name"):
            map_id = groups["name"]
            map_tag = groups.get("tag") or map_id[:3].upper()
        elif groups.get("tag"):
            map_id = groups["tag"]
            map_tag = map_id
        else:  # pragma: no cover
            raise exc

        game_mode = LAYER_GAME_MODE_MAP.get(groups["game_mode"].lower())
        if game_mode is None:
            raise exc

        if groups.get("attackers"):
            attackers = LAYER_ATTACKERS_MAP.get(groups["attackers"].lower())
        else:
            attackers = None

        time_of_day, weather = LAYER_ENVIRONMENT_MAP.get(
            groups["environment"].lower(),
            (TimeOfDay.DAY, Weather.CLEAR),
        )

        try:
            map_ = Map.by_id(map_id)
        except ValueError:
            map_ = Map(
                id=map_id,
                name=map_id.capitalize(),
                pretty_name=map_id.capitalize(),
                short_name=map_id.capitalize(),
                tag=map_tag,
                allies=(
                    attackers
                    if attackers and attackers.team == Team.ALLIES
                    else Faction.US
                ),
                axis=(
                    attackers
                    if attackers and attackers.team == Team.AXIS
                    else Faction.GER
                ),
                orientation=Orientation.HORIZONTAL,
                mirror_factions=False,
            )

        strongpoint = Strongpoint(
            name="Sector",
            center=(0.0, -9999.0, 0.0),
            radius=0.0,
        )

        if game_mode.is_large():
            grid = Grid.large(
                scale=19840,
            )
            sectors = Sector.large_layout(
                map_.orientation,
                (
                    (
                        strongpoint.model_copy(update={"name": "Sector 1A"}),
                        strongpoint.model_copy(update={"name": "Sector 1B"}),
                        strongpoint.model_copy(update={"name": "Sector 1C"}),
                    ),
                    (
                        strongpoint.model_copy(update={"name": "Sector 2A"}),
                        strongpoint.model_copy(update={"name": "Sector 2B"}),
                        strongpoint.model_copy(update={"name": "Sector 2C"}),
                    ),
                    (
                        strongpoint.model_copy(update={"name": "Sector 3A"}),
                        strongpoint.model_copy(update={"name": "Sector 3B"}),
                        strongpoint.model_copy(update={"name": "Sector 3C"}),
                    ),
                    (
                        strongpoint.model_copy(update={"name": "Sector 4A"}),
                        strongpoint.model_copy(update={"name": "Sector 4B"}),
                        strongpoint.model_copy(update={"name": "Sector 4C"}),
                    ),
                    (
                        strongpoint.model_copy(update={"name": "Sector 5A"}),
                        strongpoint.model_copy(update={"name": "Sector 5B"}),
                        strongpoint.model_copy(update={"name": "Sector 5C"}),
                    ),
                ),
            )
        else:
            grid = Grid.small()
            sectors = Sector.skirmish_layout(
                map_.orientation,
                strongpoint,
            )

        return Layer(
            id=id_,
            map=map_,
            game_mode=game_mode,
            time_of_day=time_of_day,
            weather=weather,
            grid=grid,
            sectors=sectors,
            attacking_team=attackers.team if attackers else None,
        )

    @classmethod
    def by_id(cls, id_: str, *, strict: bool = True) -> "Layer":
        """Look up a layer by its identifier.

        Parameters
        ----------
        id_ : str
            The identifier of the layer to look up.
        strict : bool, optional
            Whether to raise an exception if no such layer is known. If set to `False`,
            will attempt to generate a fallback value based on the ID. By default
            `True`.

        Returns
        -------
        Layer
            The layer with the given identifier.

        Raises
        ------
        ValueError
            If no layer with the given identifier exists.
        ValueError
            No reasonable fallback value could be generated.

        """
        try:
            return super().by_id(id_)
        except ValueError:
            if strict:
                raise

            return cls._parse_id(id_)

    @computed_field
    @cached_property
    def pretty_name(self) -> str:
        out = self.map.pretty_name
        if self.game_mode == GameMode.OFFENSIVE:
            out += " Off."
            if self.attacking_faction:
                out += f" {self.attacking_faction.short_name}"
        else:
            out += f" {self.game_mode.id.capitalize()}"

        environment: list[str] = []

        if self.time_of_day != TimeOfDay.DAY:
            environment.append(self.time_of_day.value.title())
        if self.weather != Weather.CLEAR:
            environment.append(self.weather.value.title())

        if environment:
            out += f" ({', '.join(environment)})"

        return out

    @computed_field
    @cached_property
    def attacking_faction(self) -> Faction | None:
        if self.attacking_team == Team.ALLIES:
            return self.map.allies
        if self.attacking_team == Team.AXIS:
            return self.map.axis
        return None

    @computed_field
    @cached_property
    def defending_team(self) -> Team | None:
        if self.attacking_team == Team.ALLIES:
            return Team.AXIS
        if self.attacking_team == Team.AXIS:
            return Team.ALLIES
        return None

    @computed_field
    @cached_property
    def defending_faction(self) -> Faction | None:
        if self.attacking_team == Team.ALLIES:
            return self.map.axis
        if self.attacking_team == Team.AXIS:
            return self.map.allies
        return None

    @class_cached_property
    @classmethod
    def CARENTAN_WARFARE(cls) -> "Layer":
        return cls(
            id="carentan_warfare",
            map=Map.CARENTAN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=20160,
            ),
            sectors=SECTORS_CARENTAN_LARGE,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="carentan_warfare_night",
            map=Map.CARENTAN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=20160,
            ),
            sectors=SECTORS_CARENTAN_LARGE,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="carentan_offensive_us",
            map=Map.CARENTAN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=20160,
            ),
            sectors=SECTORS_CARENTAN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="carentan_offensive_ger",
            map=Map.CARENTAN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=20160,
            ),
            sectors=SECTORS_CARENTAN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="CAR_S_1944_Day_P_Skirmish",
            map=Map.CARENTAN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(150.0, -110.0),
            ),
            sectors=SECTORS_CARENTAN_SMALL,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_SKIRMISH_RAIN(cls) -> "Layer":
        return cls(
            id="CAR_S_1944_Rain_P_Skirmish",
            map=Map.CARENTAN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.RAIN,
            grid=Grid.small(
                offset=(150.0, -110.0),
            ),
            sectors=SECTORS_CARENTAN_SMALL,
        )

    @class_cached_property
    @classmethod
    def CARENTAN_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="CAR_S_1944_Dusk_P_Skirmish",
            map=Map.CARENTAN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(150.0, -110.0),
            ),
            sectors=SECTORS_CARENTAN_SMALL,
        )

    @class_cached_property
    @classmethod
    def DRIEL_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="driel_warfare",
            map=Map.DRIEL,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_DRIEL_LARGE,
        )

    @class_cached_property
    @classmethod
    def DRIEL_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="driel_warfare_night",
            map=Map.DRIEL,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_DRIEL_LARGE,
        )

    @class_cached_property
    @classmethod
    def DRIEL_OFFENSIVE_CW_DAY(cls) -> "Layer":
        return cls(
            id="driel_offensive_us",
            map=Map.DRIEL,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_DRIEL_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def DRIEL_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="driel_offensive_ger",
            map=Map.DRIEL,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_DRIEL_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def DRIEL_SKIRMISH_DAWN(cls) -> "Layer":
        return cls(
            id="DRL_S_1944_P_Skirmish",
            map=Map.DRIEL,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(-20.0, 28190.0),
            ),
            sectors=SECTORS_DRIEL_SMALL,
        )

    @class_cached_property
    @classmethod
    def DRIEL_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="DRL_S_1944_Night_P_Skirmish",
            map=Map.DRIEL,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(-20.0, 28190.0),
            ),
            sectors=SECTORS_DRIEL_SMALL,
        )

    @class_cached_property
    @classmethod
    def DRIEL_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="DRL_S_1944_Day_P_Skirmish",
            map=Map.DRIEL,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(-20.0, 28190.0),
            ),
            sectors=SECTORS_DRIEL_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="elalamein_warfare",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_ELALAMEIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_WARFARE_DUSK(cls) -> "Layer":
        return cls(
            id="elalamein_warfare_night",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_ELALAMEIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_OFFENSIVE_B8A_DAY(cls) -> "Layer":
        return cls(
            id="elalamein_offensive_CW",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_ELALAMEIN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_OFFENSIVE_DAK_DAY(cls) -> "Layer":
        return cls(
            id="elalamein_offensive_ger",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_ELALAMEIN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="ELA_S_1942_P_Skirmish",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(-7500.0, 0.0),
            ),
            sectors=SECTORS_ELALAMEIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELALAMEIN_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="ELA_S_1942_Night_P_Skirmish",
            map=Map.EL_ALAMEIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(-7500.0, 0.0),
            ),
            sectors=SECTORS_ELALAMEIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="elsenbornridge_warfare_day",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_WARFARE_DAWN(cls) -> "Layer":
        return cls(
            id="elsenbornridge_warfare_morning",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_WARFARE_DUSK(cls) -> "Layer":
        return cls(
            id="elsenbornridge_warfare_evening",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="elsenbornridge_warfare_night",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveUS_day",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_US_DAWN(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveUS_morning",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_US_DUSK(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveUS_evening",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_US_NIGHT(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveUS_night",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveger_day",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_GER_DAWN(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveger_morning",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_GER_DUSK(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveger_evening",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_OFFENSIVE_GER_NIGHT(cls) -> "Layer":
        return cls(
            id="elsenbornridge_offensiveger_night",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.SNOW,
            grid=Grid.large(),
            sectors=SECTORS_ELSENBORNRIDGE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="elsenbornridge_skirmish_day",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.SNOW,
            grid=Grid.small(),
            sectors=SECTORS_ELSENBORNRIDGE_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_SKIRMISH_DAWN(cls) -> "Layer":
        return cls(
            id="elsenbornridge_skirmish_morning",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.SNOW,
            grid=Grid.small(),
            sectors=SECTORS_ELSENBORNRIDGE_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="elsenbornridge_skirmish_evening",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.SNOW,
            grid=Grid.small(),
            sectors=SECTORS_ELSENBORNRIDGE_SMALL,
        )

    @class_cached_property
    @classmethod
    def ELSENBORNRIDGE_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="elsenbornridge_skirmish_night",
            map=Map.ELSENBORN_RIDGE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.SNOW,
            grid=Grid.small(),
            sectors=SECTORS_ELSENBORNRIDGE_SMALL,
        )

    @class_cached_property
    @classmethod
    def FOY_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="foy_warfare",
            map=Map.FOY,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_FOY_LARGE,
        )

    @class_cached_property
    @classmethod
    def FOY_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="foy_warfare_night",
            map=Map.FOY,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_FOY_LARGE,
        )

    @class_cached_property
    @classmethod
    def FOY_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="foy_offensive_us",
            map=Map.FOY,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_FOY_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def FOY_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="foy_offensive_ger",
            map=Map.FOY,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_FOY_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def HURTGENFOREST_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="hurtgenforest_warfare_V2",
            map=Map.HURTGEN_FOREST,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HURTGENFOREST_LARGE,
        )

    @class_cached_property
    @classmethod
    def HURTGENFOREST_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="hurtgenforest_warfare_V2_night",
            map=Map.HURTGEN_FOREST,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HURTGENFOREST_LARGE,
        )

    @class_cached_property
    @classmethod
    def HURTGENFOREST_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="hurtgenforest_offensive_US",
            map=Map.HURTGEN_FOREST,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HURTGENFOREST_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def HURTGENFOREST_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="hurtgenforest_offensive_ger",
            map=Map.HURTGEN_FOREST,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HURTGENFOREST_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def HILL400_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="hill400_warfare",
            map=Map.HILL_400,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HILL400_LARGE,
        )

    @class_cached_property
    @classmethod
    def HILL400_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="hill400_warfare_night",
            map=Map.HILL_400,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HILL400_LARGE,
        )

    @class_cached_property
    @classmethod
    def HILL400_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="hill400_offensive_US",
            map=Map.HILL_400,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HILL400_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def HILL400_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="hill400_offensive_ger",
            map=Map.HILL_400,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_HILL400_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def HILL400_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="HIL_S_1944_Day_P_Skirmish",
            map=Map.HILL_400,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_HILL400_SMALL,
        )

    @class_cached_property
    @classmethod
    def HILL400_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="HIL_S_1944_Dusk_P_Skirmish",
            map=Map.HILL_400,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_HILL400_SMALL,
        )

    @class_cached_property
    @classmethod
    def HILL400_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="HIL_S_1944_Night_P_Skirmish",
            map=Map.HILL_400,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_HILL400_SMALL,
        )

    @class_cached_property
    @classmethod
    def KHARKOV_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="kharkov_warfare",
            map=Map.KHARKOV,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KHARKOV_LARGE,
        )

    @class_cached_property
    @classmethod
    def KHARKOV_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="kharkov_warfare_night",
            map=Map.KHARKOV,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KHARKOV_LARGE,
        )

    @class_cached_property
    @classmethod
    def KHARKOV_OFFENSIVE_SOV_DAY(cls) -> "Layer":
        return cls(
            id="kharkov_offensive_rus",
            map=Map.KHARKOV,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KHARKOV_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def KHARKOV_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="kharkov_offensive_ger",
            map=Map.KHARKOV,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KHARKOV_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def KURSK_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="kursk_warfare",
            map=Map.KURSK,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KURSK_LARGE,
        )

    @class_cached_property
    @classmethod
    def KURSK_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="kursk_warfare_night",
            map=Map.KURSK,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KURSK_LARGE,
        )

    @class_cached_property
    @classmethod
    def KURSK_OFFENSIVE_SOV_DAY(cls) -> "Layer":
        return cls(
            id="kursk_offensive_rus",
            map=Map.KURSK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KURSK_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def KURSK_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="kursk_offensive_ger",
            map=Map.KURSK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_KURSK_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="mortain_warfare_day",
            map=Map.MORTAIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_WARFARE_DUSK(cls) -> "Layer":
        return cls(
            id="mortain_warfare_dusk",
            map=Map.MORTAIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_WARFARE_OVERCAST(cls) -> "Layer":
        return cls(
            id="mortain_warfare_overcast",
            map=Map.MORTAIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="mortain_warfare_night",
            map=Map.MORTAIN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="mortain_offensiveUS_day",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_US_OVERCAST(cls) -> "Layer":
        return cls(
            id="mortain_offensiveUS_overcast",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_US_DUSK(cls) -> "Layer":
        return cls(
            id="mortain_offensiveUS_dusk",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_US_NIGHT(cls) -> "Layer":
        return cls(
            id="mortain_offensiveUS_night",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="mortain_offensiveger_day",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_GER_OVERCAST(cls) -> "Layer":
        return cls(
            id="mortain_offensiveger_overcast",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_GER_DUSK(cls) -> "Layer":
        return cls(
            id="mortain_offensiveger_dusk",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_OFFENSIVE_GER_NIGHT(cls) -> "Layer":
        return cls(
            id="mortain_offensiveger_night",
            map=Map.MORTAIN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_MORTAIN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="mortain_skirmish_day",
            map=Map.MORTAIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(100.0, 0.0),
            ),
            sectors=SECTORS_MORTAIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_SKIRMISH_OVERCAST(cls) -> "Layer":
        return cls(
            id="mortain_skirmish_overcast",
            map=Map.MORTAIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.small(
                offset=(100.0, 0.0),
            ),
            sectors=SECTORS_MORTAIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="mortain_skirmish_dusk",
            map=Map.MORTAIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(100.0, 0.0),
            ),
            sectors=SECTORS_MORTAIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def MORTAIN_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="mortain_skirmish_night",
            map=Map.MORTAIN,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(100.0, 0.0),
            ),
            sectors=SECTORS_MORTAIN_SMALL,
        )

    @class_cached_property
    @classmethod
    def OMAHABEACH_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="omahabeach_warfare",
            map=Map.OMAHA_BEACH,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_OMAHABEACH_LARGE,
        )

    @class_cached_property
    @classmethod
    def OMAHABEACH_WARFARE_DUSK(cls) -> "Layer":
        return cls(
            id="omahabeach_warfare_night",
            map=Map.OMAHA_BEACH,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_OMAHABEACH_LARGE,
        )

    @class_cached_property
    @classmethod
    def OMAHABEACH_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="omahabeach_offensive_us",
            map=Map.OMAHA_BEACH,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_OMAHABEACH_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def OMAHABEACH_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="omahabeach_offensive_ger",
            map=Map.OMAHA_BEACH,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_OMAHABEACH_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="PHL_L_1944_Warfare",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_PURPLEHEARTLANE_LARGE,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="PHL_L_1944_Warfare_Night",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_PURPLEHEARTLANE_LARGE,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="PHL_L_1944_OffensiveUS",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_PURPLEHEARTLANE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="PHL_L_1944_OffensiveGER",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_PURPLEHEARTLANE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_SKIRMISH_RAIN(cls) -> "Layer":
        return cls(
            id="PHL_S_1944_Rain_P_Skirmish",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.RAIN,
            grid=Grid.small(),
            sectors=SECTORS_PURPLEHEARTLANE_SMALL,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_SKIRMISH_DAWN(cls) -> "Layer":
        return cls(
            id="PHL_S_1944_Morning_P_Skirmish",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_PURPLEHEARTLANE_SMALL,
        )

    @class_cached_property
    @classmethod
    def PURPLEHEARTLANE_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="PHL_S_1944_Night_P_Skirmish",
            map=Map.PURPLE_HEART_LANE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_PURPLEHEARTLANE_SMALL,
        )

    @class_cached_property
    @classmethod
    def REMAGEN_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="remagen_warfare",
            map=Map.REMAGEN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_REMAGEN_LARGE,
        )

    @class_cached_property
    @classmethod
    def REMAGEN_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="remagen_warfare_night",
            map=Map.REMAGEN,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_REMAGEN_LARGE,
        )

    @class_cached_property
    @classmethod
    def REMAGEN_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="remagen_offensive_us",
            map=Map.REMAGEN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_REMAGEN_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def REMAGEN_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="remagen_offensive_ger",
            map=Map.REMAGEN,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_REMAGEN_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="STA_L_1942_Warfare",
            map=Map.STALINGRAD,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STALINGRAD_LARGE,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="STA_L_1942_Warfare_Night",
            map=Map.STALINGRAD,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STALINGRAD_LARGE,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_OFFENSIVE_SOV_DAY(cls) -> "Layer":
        return cls(
            id="STA_L_1942_OffensiveRUS",
            map=Map.STALINGRAD,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STALINGRAD_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="STA_L_1942_OffensiveGER",
            map=Map.STALINGRAD,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STALINGRAD_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_SKIRMISH_EVENING(cls) -> "Layer":
        return cls(
            id="STA_S_1942_P_Skirmish_Dusk",
            map=Map.STALINGRAD,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(150.0, -110.0),
            ),
            sectors=SECTORS_STALINGRAD_SMALL,
        )

    @class_cached_property
    @classmethod
    def STALINGRAD_SKIRMISH_OVERCAST(cls) -> "Layer":
        return cls(
            id="STA_S_1942_P_Skirmish_Overcast",
            map=Map.STALINGRAD,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            grid=Grid.small(
                offset=(150.0, -110.0),
            ),
            sectors=SECTORS_STALINGRAD_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="stmariedumont_warfare",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMARIEDUMONT_LARGE,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="stmariedumont_warfare_night",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMARIEDUMONT_LARGE,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="stmariedumont_off_us",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMARIEDUMONT_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="stmariedumont_off_ger",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMARIEDUMONT_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="SMDM_S_1944_Day_P_Skirmish",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(0.0, -27852.8),
            ),
            sectors=SECTORS_STMARIEDUMONT_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="SMDM_S_1944_Night_P_Skirmish",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(
                offset=(0.0, -27852.8),
            ),
            sectors=SECTORS_STMARIEDUMONT_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMARIEDUMONT_SKIRMISH_RAIN(cls) -> "Layer":
        return cls(
            id="SMDM_S_1944_Rain_P_Skirmish",
            map=Map.ST_MARIE_DU_MONT,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.RAIN,
            grid=Grid.small(
                offset=(0.0, -27852.8),
            ),
            sectors=SECTORS_STMARIEDUMONT_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="stmereeglise_warfare",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMEREEGLISE_LARGE,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="stmereeglise_warfare_night",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMEREEGLISE_LARGE,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="stmereeglise_offensive_us",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMEREEGLISE_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="stmereeglise_offensive_ger",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_STMEREEGLISE_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="SME_S_1944_Day_P_Skirmish",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_STMEREEGLISE_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_SKIRMISH_DAWN(cls) -> "Layer":
        return cls(
            id="SME_S_1944_Morning_P_Skirmish",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_STMEREEGLISE_SMALL,
        )

    @class_cached_property
    @classmethod
    def STMEREEGLISE_SKIRMISH_NIGHT(cls) -> "Layer":
        return cls(
            id="SME_S_1944_Night_P_Skirmish",
            map=Map.ST_MERE_EGLISE,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_STMEREEGLISE_SMALL,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="tobruk_warfare_day",
            map=Map.TOBRUK,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_WARFARE_DUSK(cls) -> "Layer":
        return cls(
            id="tobruk_warfare_dusk",
            map=Map.TOBRUK,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_WARFARE_DAWN(cls) -> "Layer":
        return cls(
            id="tobruk_warfare_morning",
            map=Map.TOBRUK,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_B8A_DAY(cls) -> "Layer":
        return cls(
            id="tobruk_offensivebritish_day",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_DAK_DAY(cls) -> "Layer":
        return cls(
            id="tobruk_offensiveger_day",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_B8A_DUSK(cls) -> "Layer":
        return cls(
            id="tobruk_offensivebritish_dusk",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_DAK_DUSK(cls) -> "Layer":
        return cls(
            id="tobruk_offensiveger_dusk",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_B8A_DAWN(cls) -> "Layer":
        return cls(
            id="tobruk_offensivebritish_morning",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_OFFENSIVE_DAK_DAWN(cls) -> "Layer":
        return cls(
            id="tobruk_offensiveger_morning",
            map=Map.TOBRUK,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.large(),
            sectors=SECTORS_TOBRUK_LARGE,
            attacking_team=Team.AXIS,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_SKIRMISH_DAY(cls) -> "Layer":
        return cls(
            id="tobruk_skirmish_day",
            map=Map.TOBRUK,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_TOBRUK_SMALL,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_SKIRMISH_DUSK(cls) -> "Layer":
        return cls(
            id="tobruk_skirmish_dusk",
            map=Map.TOBRUK,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DUSK,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_TOBRUK_SMALL,
        )

    @class_cached_property
    @classmethod
    def TOBRUK_SKIRMISH_DAWN(cls) -> "Layer":
        return cls(
            id="tobruk_skirmish_morning",
            map=Map.TOBRUK,
            game_mode=GameMode.SKIRMISH,
            time_of_day=TimeOfDay.DAWN,
            weather=Weather.CLEAR,
            grid=Grid.small(),
            sectors=SECTORS_TOBRUK_SMALL,
        )

    @class_cached_property
    @classmethod
    def UTAH_WARFARE_DAY(cls) -> "Layer":
        return cls(
            id="utahbeach_warfare",
            map=Map.UTAH_BEACH,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_UTAHBEACH_LARGE,
        )

    @class_cached_property
    @classmethod
    def UTAH_WARFARE_NIGHT(cls) -> "Layer":
        return cls(
            id="utahbeach_warfare_night",
            map=Map.UTAH_BEACH,
            game_mode=GameMode.WARFARE,
            time_of_day=TimeOfDay.NIGHT,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_UTAHBEACH_LARGE,
        )

    @class_cached_property
    @classmethod
    def UTAH_OFFENSIVE_US_DAY(cls) -> "Layer":
        return cls(
            id="utahbeach_offensive_us",
            map=Map.UTAH_BEACH,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_UTAHBEACH_LARGE,
            attacking_team=Team.ALLIES,
        )

    @class_cached_property
    @classmethod
    def UTAH_OFFENSIVE_GER_DAY(cls) -> "Layer":
        return cls(
            id="utahbeach_offensive_ger",
            map=Map.UTAH_BEACH,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.CLEAR,
            grid=Grid.large(
                scale=19840,
            ),
            sectors=SECTORS_UTAHBEACH_LARGE,
            attacking_team=Team.AXIS,
        )



import asyncio
from typing import Any
from unittest import mock

import pytest
from hllrcon.connection import RconConnection
from hllrcon.exceptions import HLLError
from hllrcon.pooled.rcon import PooledRcon
from hllrcon.pooled.worker import PooledRconWorker


@pytest.fixture
def pool() -> PooledRcon:
    pool = mock.Mock(spec=PooledRcon)
    pool._workers = mock.Mock(list)
    return pool


@pytest.fixture
def connection() -> RconConnection:
    mock_connection = mock.Mock(spec=RconConnection)
    mock_connection.is_connected.return_value = True
    return mock_connection


@pytest.fixture
def worker(
    monkeypatch: pytest.MonkeyPatch,
    pool: PooledRcon,
    connection: RconConnection,
) -> PooledRconWorker:
    """Fixture to create a mock PooledWorker."""

    async def get_connection(*_args: Any, **_kwargs: dict[str, Any]) -> RconConnection:  # noqa: ANN401
        return connection

    monkeypatch.setattr(RconConnection, "connect", get_connection)

    return PooledRconWorker(
        host="localhost",
        port=1234,
        password="password",
        pool=pool,
    )


@pytest.mark.asyncio
async def test_get_connection_new(
    worker: PooledRconWorker,
    connection: RconConnection,
) -> None:
    assert await worker._get_connection() == connection


@pytest.mark.asyncio
async def test_get_connection_failure(
    monkeypatch: pytest.MonkeyPatch,
    worker: PooledRconWorker,
) -> None:
    async def get_connection(*_args: Any, **_kwargs: dict[str, Any]) -> RconConnection:  # noqa: ANN401
        msg = "Connection failed"
        raise HLLError(msg)

    monkeypatch.setattr("hllrcon.rcon.RconConnection.connect", get_connection)

    with pytest.raises(HLLError, match="Connection failed"):
        await worker.wait_until_connected()


@pytest.mark.asyncio
async def test_get_connection_reuse(
    worker: PooledRconWorker,
    connection: RconConnection,
) -> None:
    worker._connection = asyncio.Future()
    worker._connection.set_result(connection)
    assert await worker._get_connection() == connection


@pytest.mark.asyncio
async def test_get_connection_reuse_failure(
    worker: PooledRconWorker,
) -> None:
    worker._connection = asyncio.Future()
    worker._connection.set_exception(HLLError("Connection failed"))

    with pytest.raises(HLLError, match="Connection failed"):
        await worker.wait_until_connected()


@pytest.mark.asyncio
async def test_get_connection_wait(
    worker: PooledRconWorker,
    connection: mock.Mock,
) -> None:
    worker._connection = asyncio.Future()

    with pytest.raises(asyncio.TimeoutError):
        async with asyncio.timeout(0.1):
            await worker.wait_until_connected()

    asyncio.get_event_loop().call_soon(worker._connection.set_result, connection)
    async with asyncio.timeout(0.1):
        assert await worker._get_connection() == connection


def test_is_busy(worker: PooledRconWorker) -> None:
    assert not worker.is_busy()
    worker._busy = True
    assert worker.is_busy()
    worker._busy = False
    assert not worker.is_busy()


@pytest.mark.asyncio
async def test_is_connected(worker: PooledRconWorker, connection: mock.Mock) -> None:
    assert worker.is_connected() is False, "Should be disconnected initially"

    await worker.wait_until_connected()
    assert worker.is_connected() is True, "Should be connected after getting connection"

    connection.is_connected.return_value = False
    assert worker.is_connected() is False, (
        "Should be disconnected after connection loss"
    )

    worker._connection = asyncio.Future()
    worker._connection.set_exception(HLLError("Connection lost"))
    assert worker.is_connected() is False, "Should be disconnected after exception"

    worker._connection = asyncio.Future()
    worker._connection.cancel()
    assert worker.is_connected() is False, "Should be disconnected after cancelling"


def test_is_disconnected(worker: PooledRconWorker) -> None:
    assert worker.is_disconnected() is False
    worker._disconnected = True
    assert worker.is_disconnected() is True
    worker._disconnected = False
    assert worker.is_disconnected() is False


@pytest.mark.asyncio
async def test_on_disconnected(
    worker: PooledRconWorker,
    connection: RconConnection,
) -> None:
    await worker.wait_until_connected()

    assert worker.is_disconnected() is False
    worker._busy = True

    connection.on_disconnect()
    assert worker.is_disconnected() is True
    assert worker.is_busy() is False


@pytest.mark.asyncio
async def test_disconnect_while_connected(
    worker: PooledRconWorker,
    connection: mock.Mock,
) -> None:
    await worker.wait_until_connected()

    assert not worker.is_disconnected()
    assert not worker.is_busy()

    worker.disconnect()

    assert worker.is_disconnected()
    assert not worker.is_busy()

    # Ensure the connection's disconnect method was called
    connection.disconnect.assert_called_once()


@pytest.mark.asyncio
async def test_disconnect_before_connecting(
    worker: PooledRconWorker,
) -> None:
    assert not worker.is_disconnected()
    assert not worker.is_busy()

    worker.disconnect()

    assert worker.is_disconnected()
    assert not worker.is_busy()


@pytest.mark.asyncio
async def test_disconnect_while_connecting(
    worker: PooledRconWorker,
) -> None:
    fut: asyncio.Future[RconConnection] = asyncio.Future()
    worker._connection = fut

    assert not worker.is_disconnected()
    assert not worker.is_busy()

    worker.disconnect()

    assert worker.is_disconnected()
    assert not worker.is_busy()

    # Ensure that the connection attempt was cancelled
    assert fut.cancelled() is True


@pytest.mark.asyncio
async def test_disconnect_while_exception(
    worker: PooledRconWorker,
) -> None:
    fut: asyncio.Future[RconConnection] = asyncio.Future()
    fut.set_exception(HLLError("Connection failed"))
    worker._connection = fut

    assert not worker.is_disconnected()
    assert not worker.is_busy()

    worker.disconnect()

    assert worker.is_disconnected()
    assert not worker.is_busy()


@pytest.mark.asyncio
async def test_disconnect_while_cancelled(
    worker: PooledRconWorker,
) -> None:
    fut: asyncio.Future[RconConnection] = asyncio.Future()
    fut.cancel()
    worker._connection = fut

    assert not worker.is_disconnected()
    assert not worker.is_busy()

    worker.disconnect()

    assert worker.is_disconnected()
    assert not worker.is_busy()


@pytest.mark.asyncio
async def test_execute(
    worker: PooledRconWorker,
    connection: mock.Mock,
) -> None:
    command = "test_command"
    version = 1
    body = "test_body"

    # Mock the execute method of the connection
    connection.execute = mock.AsyncMock(return_value="response")

    response = await worker.execute(command, version, body)

    assert response == "response"
    connection.execute.assert_awaited_once_with(command, version, body)



from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Any, cast

import pytest
from hllrcon.data import GameMode, Layer, Role
from hllrcon.responses import (
    EmptyStringToNoneValidator,
    GetAdminLogResponseEntry,
    GetMapRotationResponseEntry,
    GetPlayerResponse,
    GetServerSessionResponse,
    PlayerFactionId,
    PlayerPlatform,
    PlayerRoleId,
)

if TYPE_CHECKING:
    from collections.abc import Callable


def test_empty_string_to_none_validator() -> None:
    validator_func = cast("Callable[[Any], Any]", EmptyStringToNoneValidator.func)

    # Test with an empty string
    assert validator_func("") is None

    # Test with a non-empty string
    assert validator_func("Test") == "Test"

    # Test with None
    assert validator_func(None) is None


def test_get_admin_log_response_convert_isoformat_to_datetime() -> None:
    entry = GetAdminLogResponseEntry.model_validate(
        {
            "timestamp": "2023-03-15T12:34:56.789Z",
            "message": "Test log entry",
        },
    )
    assert entry.timestamp == datetime(2023, 3, 15, 12, 34, 56, 789000, tzinfo=UTC)


def test_get_player_response_team_unassigned() -> None:
    response = GetPlayerResponse.model_validate(
        {
            "name": "TestPlayer",
            "clanTag": "",
            "iD": "12345",
            "platform": PlayerPlatform.STEAM.value,
            "eosId": "eos-12345",
            "level": 10,
            "team": PlayerFactionId.UNASSIGNED.value,
            "role": PlayerRoleId.RIFLEMAN.value,
            "platoon": "",
            "loadout": "Standard Issue",
            "kills": 5,
            "deaths": 2,
            "scoreData": {
                "cOMBAT": 100,
                "offense": 50,
                "defense": 30,
                "support": 20,
            },
            "worldPosition": {
                "x": 0.0,
                "y": 0.0,
                "z": 0.0,
            },
        },
    )

    assert response.faction is None
    assert response.role is Role.RIFLEMAN

    # Test other validators just for the sake of it
    assert response.clan_tag is None
    assert response.platoon is None


def test_get_server_session_game_mode() -> None:
    response = GetServerSessionResponse(
        server_name="Test Server",
        map_name="FOY",
        game_mode_id="Warfare",
        remaining_match_time=timedelta(seconds=300),
        match_time=600,
        allied_score=2,
        axis_score=3,
        player_count=64,
        allied_player_count=32,
        axis_player_count=32,
        max_player_count=100,
        queue_count=0,
        max_queue_count=6,
        vip_queue_count=0,
        max_vip_queue_count=6,
    )
    assert response.game_mode == GameMode.WARFARE


def test_map_rotation_entry_find_layer() -> None:
    entry = GetMapRotationResponseEntry(
        name="FOY",
        game_mode_name="Warfare",
        time_of_day="DAY",
        id="foy_warfare",
        position=0,
    )
    assert entry.game_mode == GameMode.WARFARE
    assert entry.find_layer() == Layer.FOY_WARFARE_DAY

    entry.game_mode_name = "Control Skirmish - La Petite Chapelle"
    assert entry.game_mode == GameMode.SKIRMISH

    unknown_entry = GetMapRotationResponseEntry(
        name="SOY",
        game_mode_name="Offensive",
        time_of_day="DAY",
        id="not_foy_warfare",
        position=1,
    )
    assert unknown_entry.game_mode == GameMode.OFFENSIVE
    with pytest.raises(ValueError, match="not found"):
        unknown_entry.find_layer()



######################################################
#  Generated using `scripts/generate_sync_commands.py`
#  DO NOT EDIT THIS FILE DIRECTLY
######################################################

from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps
from typing import Any, Literal, ParamSpec, TypeVar

from pydantic import BaseModel

from hllrcon.data import layers
from hllrcon.data.factions import Faction
from hllrcon.exceptions import HLLCommandError, HLLMessageError
from hllrcon.responses import (
    ForceMode,
    GetAdminGroupsResponse,
    GetAdminLogResponse,
    GetAdminUsersResponse,
    GetBannedWordsResponse,
    GetBansResponse,
    GetCommandDetailsResponse,
    GetCommandsResponse,
    GetMapRotationResponse,
    GetPlayerResponse,
    GetPlayersResponse,
    GetServerConfigResponse,
    GetServerSessionResponse,
    GetVipsResponse,
    PlayerFactionId,
)

P = ParamSpec("P")
ModelT = TypeVar("ModelT", bound=BaseModel)

GameMode = Literal["Warfare", "Offensive", "Skirmish"]


def cast_response_to_model(
    model_type: type[ModelT],
) -> Callable[
    [Callable[P, str]],
    Callable[P, ModelT],
]:
    def decorator(
        func: Callable[P, str],
    ) -> Callable[P, ModelT]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> ModelT:
            result = func(*args, **kwargs)
            return model_type.model_validate_json(result)

        return wrapper

    return decorator


def cast_response_to_bool(
    status_codes: set[int],
) -> Callable[
    [Callable[P, None]],
    Callable[P, bool],
]:
    def decorator(
        func: Callable[P, None],
    ) -> Callable[P, bool]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> bool:
            try:
                func(*args, **kwargs)
            except HLLCommandError as e:
                if e.status_code in status_codes:
                    return False
                raise
            return True

        return wrapper

    return decorator


class SyncRconCommands(ABC):
    @abstractmethod
    def execute(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
    ) -> str:
        """Execute a command on the RCON server.

        Parameters
        ----------
        command : str
            The command to execute.
        version : int
            The version of the command to execute.
        body : str | dict[str, Any], optional
            The body of the command, by default an empty string.

        Returns
        -------
        str
            The response from the server.

        """

    def add_admin(self, player_id: str, admin_group: str, comment: str) -> None:
        """Add a player to an admin group.

        Groups are defined in the server's configuration file. The group determines
        whether the player is able to enter admin camera and kick or ban players.

        Parameters
        ----------
        player_id : str
            The ID of the player to add as an admin.
        admin_group : str
            The group to add the player to.
        comment : str
            A comment to identify the admin. This is usually the name of the player.

        """
        self.execute(
            "AddAdmin",
            2,
            {"PlayerId": player_id, "AdminGroup": admin_group, "Comment": comment},
        )

    def remove_admin(self, player_id: str) -> None:
        """Remove a player from their admin group.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove as an admin.

        """
        self.execute(
            "RemoveAdmin",
            2,
            {
                "PlayerId": player_id,
            },
        )

    @cast_response_to_model(GetAdminLogResponse)
    def get_admin_log(self, seconds_span: int, filter_: str | None = None) -> str:
        """Retrieve admin logs from the server.

        Parameters
        ----------
        seconds_span : int
            The number of seconds to look back in the logs.
        filter_ : str | None
            A filter string to apply to the logs, by default None.

        Returns
        -------
        GetAdminLogResponse
            A response containing the admin logs.

        """
        if seconds_span < 0:
            msg = "seconds_span must be a non-negative integer"
            raise ValueError(msg)

        return self.execute(
            "GetAdminLog",
            2,
            {
                "LogBackTrackTime": seconds_span,
                "Filters": filter_ or "",
            },
        )

    def change_map(self, map_name: str | layers.Layer) -> None:
        """Change the current map to the specified map.

        Map changes are not immediate. Instead, a 60 second countdown is started.

        Parameters
        ----------
        map_name : str | Layer
            The name of the map to change to.

        """
        self.execute(
            "ChangeMap",
            2,
            {
                "MapName": str(map_name),
            },
        )

    def get_available_sector_names(
        self,
    ) -> tuple[list[str], list[str], list[str], list[str], list[str]]:
        """Retrieve a list of all sector names available on the current map.

        Returns
        -------
        tuple[list[str], list[str], list[str], list[str], list[str]]
            A list of sector names available on the current map.

        """
        details = self.get_command_details("SetSectorLayout")
        parameters = details.dialogue_parameters
        if not parameters or not all(
            p.id.startswith("Sector_") for p in parameters[:5]
        ):
            msg = "Received unexpected response from server."
            raise HLLMessageError(msg)
        return (
            parameters[0].value_member,
            parameters[1].value_member,
            parameters[2].value_member,
            parameters[3].value_member,
            parameters[4].value_member,
        )

    def set_sector_layout(
        self,
        sector1: str,
        sector2: str,
        sector3: str,
        sector4: str,
        sector5: str,
    ) -> None:
        """Immediately restart the map with the given sector layout.

        Parameters
        ----------
        sector1 : str
            The name of the first sector.
        sector2 : str
            The name of the second sector.
        sector3 : str
            The name of the third sector.
        sector4 : str
            The name of the fourth sector.
        sector5 : str
            The name of the fifth sector.

        """
        self.execute(
            "SetSectorLayout",
            2,
            {
                "Sector_1": sector1,
                "Sector_2": sector2,
                "Sector_3": sector3,
                "Sector_4": sector4,
                "Sector_5": sector5,
            },
        )

    def add_map_to_rotation(
        self,
        map_name: str | layers.Layer,
        index: int,
    ) -> None:
        """Add a map to the map rotation.

        Parameters
        ----------
        map_name : str | Layer
            The name of the map to add.
        index : int
            The index in the rotation to add the map at.

        """
        self.execute(
            "AddMapToRotation",
            2,
            {
                "MapName": str(map_name),
                "Index": index,
            },
        )

    def remove_map_from_rotation(self, index: int) -> None:
        """Remove a map from the map rotation.

        Parameters
        ----------
        index : int
            The index of the map to remove from the rotation.

        """
        self.execute(
            "RemoveMapFromRotation",
            2,
            {
                "Index": index,
            },
        )

    def add_map_to_sequence(
        self,
        map_name: str | layers.Layer,
        index: int,
    ) -> None:
        """Add a map to the map sequence.

        Parameters
        ----------
        map_name : str | Layer
            The name of the map to add.
        index : int
            The index in the sequence to add the map at.

        """
        self.execute(
            "AddMapToSequence",
            2,
            {
                "MapName": str(map_name),
                "Index": index,
            },
        )

    def remove_map_from_sequence(self, index: int) -> None:
        """Remove a map from the map sequence.

        Parameters
        ----------
        index : int
            The index of the map to remove from the sequence.

        """
        self.execute(
            "RemoveMapFromSequence",
            2,
            {
                "Index": index,
            },
        )

    def set_map_shuffle_enabled(self, *, enabled: bool) -> None:
        """Enable or disable map shuffling of the map sequence.

        Parameters
        ----------
        enabled : bool
            Whether to enable or disable map shuffling.

        """
        self.execute(
            "SetMapShuffleEnabled",
            2,
            {
                "Enable": enabled,
            },
        )

    def move_map_in_sequence(self, old_index: int, new_index: int) -> None:
        """Move a map in the map sequence.

        Parameters
        ----------
        old_index : int
            The current index of the map in the sequence.
        new_index : int
            The new index to move the map to in the sequence.

        """
        self.execute(
            "MoveMapInSequence",
            2,
            {
                "CurrentIndex": old_index,
                "NewIndex": new_index,
            },
        )

    def get_available_maps(self) -> list[str]:
        """Retrieve a list of all maps available on the server.

        Returns
        -------
        list[str]
            A list of map names available on the server.

        """
        details = self.get_command_details("AddMapToRotation")
        parameters = details.dialogue_parameters
        if not parameters or parameters[0].id != "MapName":
            msg = "Received unexpected response from server."
            raise HLLMessageError(msg)
        return parameters[0].value_member

    @cast_response_to_model(GetCommandsResponse)
    def get_commands(self) -> str:
        """Retrieve a description of all the commands available on the server.

        Returns
        -------
        GetAllCommandsResponse
            A response containing a list of all commands available on the server.

        """
        return self.execute("GetDisplayableCommands", 2)

    @cast_response_to_model(GetAdminGroupsResponse)
    def get_admin_groups(self) -> str:
        """Retrieve a list of all admin groups available on the server.

        Admin groups are defined inside of the server's `AdminPermissions.ini` file, and
        control what permissions admins have on the server, such as entering admin
        camera and kicking or banning players.

        Returns
        -------
        GetAdminGroupsResponse
            A list of all available admin groups.

        """
        return self.execute("GetAdminGroups", 2)

    @cast_response_to_model(GetAdminUsersResponse)
    def get_admin_users(self) -> str:
        """Retrieve a list of all users with admin permissions.

        Returns
        -------
        GetAdminUsersResponse
            A list of all admin users.

        """
        return self.execute("GetAdminUsers", 2)

    @cast_response_to_model(GetBansResponse)
    def get_permanent_bans(self) -> str:
        """Retrieve a list of all permanently banned players.

        Returns
        -------
        GetBansGesponse
            A list of all permanently banned players.

        """
        return self.execute("GetPermanentBans", 2)

    @cast_response_to_model(GetBansResponse)
    def get_temporary_bans(self) -> str:
        """Retrieve a list of all temporarily banned players.

        Returns
        -------
        GetBansResponse
            A list of all temporarily banned players.

        """
        return self.execute("GetTemporaryBans", 2)

    @cast_response_to_bool({400})
    def disband_squad(
        self,
        faction: Faction | PlayerFactionId | int,
        squad_index: int,
        reason: str,
    ) -> None:
        """Disband a squad on the server.

        Parameters
        ----------
        faction : Faction | PlayerFactionId | int
            The ID of the faction the squad belongs to.
        squad_index : int
            The index of the squad to disband.
        reason : str
            The reason for disbanding the squad. This will be displayed to the players.

        """
        team_index = faction.id if isinstance(faction, Faction) else int(faction)
        self.execute(
            "DisbandPlatoon",
            2,
            {
                "TeamIndex": team_index,
                "SquadIndex": squad_index,
                "Reason": reason,
            },
        )

    # TODO: Verify that responses are correctly casted to bool
    @cast_response_to_bool({500})
    def force_team_switch(
        self,
        player_id: str,
        force_mode: ForceMode = ForceMode.IMMEDIATE,
    ) -> None:
        """Force a player to switch to a specific team.

        Parameters
        ----------
        player_id : str
            The ID of the player to force switch.
        force_mode : ForceMode
            When to force the player to switch, by default `ForceMode.IMMEDIATE`.

        Returns
        -------
        bool
            Whether the player was successfully forced to switch teams.

        """
        self.execute(
            "ForceTeamSwitch",
            2,
            {
                "PlayerId": player_id,
                "ForceMode": force_mode,
            },
        )

    def set_team_switch_cooldown(self, minutes: int) -> None:
        """Set the cooldown for switching teams.

        Parameters
        ----------
        minutes : int
            The number of minutes to set the cooldown to. Set to 0 for no cooldown.

        """
        self.execute(
            "SetTeamSwitchCooldown",
            2,
            {
                "TeamSwitchTimer": minutes,
            },
        )

    def set_max_queued_players(self, num: int) -> None:
        """Set the maximum number of players that can be queued for the server.

        Parameters
        ----------
        num : int
            The maximum number of players that can be queued. Must be between 0 and 6.

        """
        self.execute(
            "SetMaxQueuedPlayers",
            2,
            {
                "MaxQueuedPlayers": num,
            },
        )

    def set_idle_kick_duration(self, minutes: int) -> None:
        """Set the number of minutes a player can be idle for before being kicked.

        Parameters
        ----------
        minutes : int
            The number of minutes a player can be idle for before being kicked.
            Set to 0 to disable.

        """
        self.execute(
            "SetIdleKickDuration",
            2,
            {
                "IdleTimeoutMinutes": minutes,
            },
        )

    def set_welcome_message(self, message: str) -> None:
        """Set the welcome message for the server.

        The welcome message is displayed to players on the deployment screen and briefly
        when they first spawn in. The message will be briefly shown again when updated.

        Parameters
        ----------
        message : str
            The welcome message to set.

        """
        self.execute(
            "SetWelcomeMessage",
            2,
            {
                "Message": message,
            },
        )

    @cast_response_to_model(GetPlayerResponse)
    def get_player(self, player_id: str) -> str:
        """Retrieve detailed information about a player currently on the server.

        Parameters
        ----------
        player_id : str
            The ID of the player to retrieve information about.

        Returns
        -------
        GetPlayerResponse
            Information about the player.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "player", "Value": player_id},
        )

    @cast_response_to_model(GetPlayersResponse)
    def get_players(self) -> str:
        """Retrieve detailed information about all players currently on the server.

        This is equivalent to calling `get_player` for each player on the server.

        Returns
        -------
        GetPlayersResponse
            Information about all players.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "players", "Value": ""},
        )

    @cast_response_to_model(GetMapRotationResponse)
    def get_map_rotation(self) -> str:
        """Retrieve the current map rotation of the server.

        Returns
        -------
        GetMapRotationResponse
            The current map rotation of the server.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "maprotation", "Value": ""},
        )

    @cast_response_to_model(GetMapRotationResponse)
    def get_map_sequence(self) -> str:
        """Retrieve the current map sequence of the server.

        Returns
        -------
        GetMapRotationResponse
            The current map sequence of the server.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "mapsequence", "Value": ""},
        )

    @cast_response_to_model(GetServerSessionResponse)
    def get_server_session(self) -> str:
        """Retrieve information abou the current server session.

        Returns
        -------
        GetServerSessionResponse
            Information about the current server session.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "session", "Value": ""},
        )

    @cast_response_to_model(GetServerConfigResponse)
    def get_server_config(self) -> str:
        """Retrieve the server configuration.

        Returns
        -------
        GetServerConfigResponse
            The server configuration.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "serverconfig", "Value": ""},
        )

    @cast_response_to_model(GetBannedWordsResponse)
    def get_banned_words(self) -> str:
        """Retrieve the list of banned words on the server.

        Returns
        -------
        GetBannedWordsResponse
            The list of banned words on the server.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "bannedwords", "Value": ""},
        )

    @cast_response_to_model(GetVipsResponse)
    def get_vip_users(self) -> str:
        """Retrieve the list of VIPs.

        Returns
        -------
        GetVipsResponse
            The list of VIPs.

        """
        return self.execute(
            "GetServerInformation",
            2,
            {"Name": "vipplayers", "Value": ""},
        )

    def broadcast(self, message: str) -> None:
        """Broadcast a message to all players on the server.

        Broadcast messages are displayed top-left on the screen for all players.

        Parameters
        ----------
        message : str
            The message to broadcast to all players.

        """
        self.execute(
            "ServerBroadcast",
            2,
            {
                "Message": message,
            },
        )

    def set_high_ping_threshold(self, ms: int) -> None:
        """Set the ping threshold for players.

        If a player's ping exceeds this threshold, they will be kicked from the server.

        Parameters
        ----------
        ms : int
            The ping threshold in milliseconds. Set to 0 to disable.

        """
        self.execute(
            "SetHighPingThreshold",
            2,
            {
                "HighPingThresholdMs": ms,
            },
        )

    @cast_response_to_model(GetCommandDetailsResponse)
    def get_command_details(self, command: str) -> str:
        """Retrieve detailed information about a specific command.

        Parameters
        ----------
        command : str
            The name of the command to retrieve information about.

        Returns
        -------
        GetCommandDetailsResponse
            Information about the command, including its parameters and description.

        """
        return self.execute("GetClientReferenceData", 2, command)

    def message_player(self, player_id: str, message: str) -> None:
        """Send a message to a specific player on the server.

        The message will be displayed in a box in the top right corner of the player's
        screen.

        Parameters
        ----------
        player_id : str
            The ID of the player to send the message to.
        message : str
            The message to send to the player.

        """
        self.execute(
            "MessagePlayer",
            2,
            {
                "Message": message,
                "PlayerId": player_id,
            },
        )

    @cast_response_to_bool({500})
    def kill_player(self, player_id: str, message: str | None = None) -> None:
        """Kill a specific player on the server.

        Parameters
        ----------
        player_id : str
            The ID of the player to kill.
        message : str | None
            The reason for killing the player. This will be displayed to the player, by
            default None.

        Returns
        -------
        bool
            Whether the player was successfully killed. If the player is not on the
            server or already dead, this will return `False`.

        """
        self.execute(
            "PunishPlayer",
            2,
            {
                "PlayerId": player_id,
                "Reason": message,
            },
        )

    @cast_response_to_bool({400})
    def kick_player(self, player_id: str, message: str) -> None:
        """Kick a specific player from the server.

        Parameters
        ----------
        player_id : str
            The ID of the player to kick.
        message : str
            The reason for kicking the player. This will be displayed to the player.

        Returns
        -------
        bool
            Whether the player was successfully kicked. If the player is not on the
            server, this will return `False`.

        """
        self.execute(
            "KickPlayer",
            2,
            {
                "PlayerId": player_id,
                "Reason": message,
            },
        )

    def ban_player(
        self,
        player_id: str,
        reason: str,
        admin_name: str,
        duration_hours: int | None = None,
    ) -> None:
        """Ban a specific player from the server.

        Parameters
        ----------
        player_id : str
            The ID of the player to ban.
        reason : str
            The reason for banning the player. This will be displayed to the player.
        admin_name : str
            The name of the admin performing the ban.
        duration_hours : int | None, optional
            The duration of the ban in hours. If `None`, the player will be permanently
            banned. Defaults to `None`.

        """
        if duration_hours:
            self.execute(
                "TemporaryBanPlayer",
                2,
                {
                    "PlayerId": player_id,
                    "Duration": duration_hours,
                    "Reason": reason,
                    "AdminName": admin_name,
                },
            )
        else:
            self.execute(
                "PermanentBanPlayer",
                2,
                {
                    "PlayerId": player_id,
                    "Reason": reason,
                    "AdminName": admin_name,
                },
            )

    @cast_response_to_bool({400})
    def remove_temporary_ban(self, player_id: str) -> None:
        """Remove a temporary ban for a specific player.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove the temporary ban for.

        Returns
        -------
        bool
            Whether the player was successfully unbanned. If the player is not
            temporarily banned, this will return `False`.

        """
        self.execute(
            "RemoveTemporaryBan",
            2,
            {
                "PlayerId": player_id,
            },
        )

    @cast_response_to_bool({400})
    def remove_permanent_ban(self, player_id: str) -> None:
        """Remove a permanent ban for a specific player.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove the permanent ban for.

        Returns
        -------
        bool
            Whether the player was successfully unbanned. If the player is not
            permanently banned, this will return `False`.

        """
        self.execute(
            "RemovePermanentBan",
            2,
            {
                "PlayerId": player_id,
            },
        )

    def unban_player(self, player_id: str) -> bool:
        """Remove any temporary or permanent ban for a specific player.

        This is equivalent to calling both `remove_temporary_ban` and
        `remove_permanent_ban` for the player.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove the ban for.

        Returns
        -------
        bool
            Whether the player was successfully unbanned. If the player is not
            banned, this will return `False`.

        """
        responses = (
            self.remove_temporary_ban(player_id),
            self.remove_permanent_ban(player_id),
        )
        return any(responses)

    @cast_response_to_bool({400})
    def remove_player_from_squad(self, player_id: str, reason: str) -> None:
        """Remove a player from their current squad.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove from their squad.
        reason : str
            The reason for removing the player from their squad. This will be displayed
            to the players.

        """
        self.execute(
            "RemovePlayerFromPlatoon",
            2,
            {
                "PlayerId": player_id,
                "Reason": reason,
            },
        )

    def set_auto_balance_enabled(self, *, enabled: bool) -> None:
        """Enable or disable team balancing.

        When enabled, the server will prevent players from joining a team when
        it has significantly more players than the other team. This threshold is
        configurable with `set_auto_balance_threshold`.

        Parameters
        ----------
        enabled : bool
            Whether to enable or disable auto balancing.

        """
        self.execute(
            "SetAutoBalanceEnabled",
            2,
            {
                "Enable": enabled,
            },
        )

    def set_auto_balance_threshold(self, player_threshold: int) -> None:
        """Set the player threshold for auto balancing.

        For example, with a threshold of 2, a player is only allowed to join a team if
        after joining it would have at most 2 more players than the other team.

        Setting the threshold to 0 will NOT disable auto balancing! For that purpose,
        use `set_auto_balance_enabled` instead.

        Parameters
        ----------
        player_threshold : int
            The number of players that can be more on one team than the other.
            Must be a non-negative integer.

        """
        self.execute(
            "SetAutoBalanceThreshold",
            2,
            {
                "AutoBalanceThreshold": player_threshold,
            },
        )

    def set_vote_kick_enabled(self, *, enabled: bool) -> None:
        """Enable or disable vote kicking.

        Parameters
        ----------
        enabled : bool
            Whether to enable or disable vote kicking.

        """
        self.execute(
            "SetVoteKickEnabled",
            2,
            {
                "Enable": enabled,
            },
        )

    def reset_vote_kick_thresholds(self) -> None:
        """Reset the vote kick thresholds to the default values."""
        self.execute("ResetVoteKickThreshold", 2)

    def set_vote_kick_thresholds(self, thresholds: list[tuple[int, int]]) -> None:
        """Set the vote kick thresholds for different player counts.

        The thresholds are a list of tuples, where each tuple contains the number of
        players and the number of votes required to kick a player. For example, a
        threshold of (5, 3) means that if there are 5 players on the server, 3 votes
        are required to kick a player.

        Parameters
        ----------
        thresholds : list[tuple[int, int]]
            A list of tuples containing the player count and the number of votes
            required to kick a player.

        """
        self.execute(
            "SetVoteKickThreshold",
            2,
            {
                "ThresholdValue": ",".join([f"{p},{v}" for p, v in thresholds]),
            },
        )

    def add_banned_words(self, words: list[str]) -> None:
        """Add words to the list of banned words on the server.

        Banned works will be replaced with asterisks when sent in chat.

        Parameters
        ----------
        words : list[str]
            A list of words or phrases to add to the list of banned words.

        """
        self.execute(
            "AddBannedWords",
            2,
            {
                "Words": ",".join(words),
            },
        )

    def remove_banned_words(self, words: list[str]) -> None:
        """Remove words from the list of banned words on the server.

        Parameters
        ----------
        words : list[str]
            A list of words or phrases to remove from the list of banned words.

        """
        self.execute(
            "RemoveBannedWords",
            2,
            {
                "Words": ",".join(words),
            },
        )

    def add_vip(self, player_id: str, description: str) -> None:
        """Add a player to the VIP list.

        Parameters
        ----------
        player_id : str
            The ID of the player to add as a VIP.
        description : str
            A description of the VIP. This is usually the name of the player.

        """
        self.execute(
            "AddVip",
            2,
            {
                "PlayerId": player_id,
                "Comment": description,
            },
        )

    def remove_vip(self, player_id: str) -> None:
        """Remove a player from the VIP list.

        Parameters
        ----------
        player_id : str
            The ID of the player to remove from the VIP list.

        """
        self.execute(
            "RemoveVip",
            2,
            {
                "PlayerId": player_id,
            },
        )

    def set_num_vip_slots(self, num: int) -> None:
        """Set the number of VIP slots on the server.

        This determines how many players can be in the VIP list at the same time.
        If the number of VIPs exceeds this limit, the oldest VIP will be removed.

        Parameters
        ----------
        num : int
            The number of VIP slots to set. Must be a non-negative integer.

        """
        self.execute(
            "SetVipSlotCount",
            2,
            {
                "VipSlotCount": num,
            },
        )

    def set_match_timer(self, game_mode: GameMode, minutes: int) -> None:
        """Set the match timer for a specific game mode.

        This does not affect the current match, but will apply to all future matches
        of the specified game mode. Limits apply, depending on the game mode.

        Parameters
        ----------
        game_mode : GameMode
            The game mode to set the match timer for. One of "Warfare", "Offensive",
            or "Skirmish".
        minutes : int
            The number of minutes to set the match timer to.

        """
        self.execute(
            "SetMatchTimer",
            2,
            {
                "GameMode": game_mode,
                "MatchLength": minutes,
            },
        )

    def reset_match_timer(self, game_mode: GameMode) -> None:
        """Reset the match timer for a specific game mode.

        This does not affect the current match, but will apply to all future matches
        of the specified game mode. The match timer will be set to the default value.

        Parameters
        ----------
        game_mode : GameMode
            The game mode to reset the match timer for. One of "Warfare", "Offensive",
            or "Skirmish".

        """
        self.execute(
            "RemoveMatchTimer",
            2,
            {
                "GameMode": game_mode,
            },
        )

    def set_warmup_timer(self, game_mode: GameMode, minutes: int) -> None:
        """Set the warmup timer for a specific game mode.

        This does not affect the current match, but will apply to all future matches
        of the specified game mode. Limits apply, depending on the game mode.

        Parameters
        ----------
        game_mode : GameMode
            The game mode to set the warmup timer for. One of "Warfare", "Offensive",
            or "Skirmish".
        minutes : int
            The number of minutes to set the warmup timer to.

        """
        self.execute(
            "SetWarmupTimer",
            2,
            {
                "GameMode": game_mode,
                "WarmupLength": minutes,
            },
        )

    def remove_warmup_timer(self, game_mode: GameMode) -> None:
        """Reset the warmup timer for a specific game mode.

        This does not affect the current match, but will apply to all future matches
        of the specified game mode. The warmup timer will be set to the default value.

        Parameters
        ----------
        game_mode : GameMode
            The game mode to reset the warmup timer for. One of "Warfare", "Offensive",
            or "Skirmish".

        """
        self.execute(
            "RemoveWarmupTimer",
            2,
            {
                "GameMode": game_mode,
            },
        )

    def set_dynamic_weather_enabled(self, map_id: str, *, enabled: bool) -> None:
        """Enable or disable dynamic weather for a specific map.

        Not all maps have dynamic weather. Maps that do not support dynamic weather
        will ignore this command.

        This does not affect the current match, but will apply to all future
        matches on the specified map.

        Parameters
        ----------
        map_id : str
            The ID of the map to enable or disable dynamic weather for.
        enabled : bool
            Whether to enable or disable dynamic weather for the map.

        """
        self.execute(
            "SetDynamicWeatherEnabled",
            2,
            {
                "MapId": map_id,
                "Enable": enabled,
            },
        )
